#ifndef CLASS_NCINE_APPLICATION
#define CLASS_NCINE_APPLICATION

#include "IGfxDevice.h"
#include "AppConfiguration.h"
#include "IDebugOverlay.h"
#include "TimeStamp.h"
#include <nctl/UniquePtr.h>

namespace ncine {

class FrameTimer;
class SceneNode;
class RenderQueue;
class IInputManager;
class IAppEventHandler;
class ImGuiDrawing;

/// Main entry point and handler for nCine applications
class DLL_PUBLIC Application
{
  public:
	/// Rendering settings that can be changed at run-time
	struct RenderingSettings
	{
		RenderingSettings()
		    : batchingEnabled(true), batchingWithIndices(false),
		      cullingEnabled(true), minBatchSize(4), maxBatchSize(500) {}

		/// True if batching is enabled
		bool batchingEnabled;
		/// True if using indices for vertex batching
		bool batchingWithIndices;
		/// True if node culling is enabled
		bool cullingEnabled;
		/// Minimum size for a batch to be collected
		unsigned int minBatchSize;
		/// Maximum size for a batch before a forced split
		unsigned int maxBatchSize;
	};

	struct Timings
	{
		enum
		{
			PRE_INIT,
			INIT_COMMON,
			APP_INIT,
			FRAME_START,
			UPDATE_VISIT_DRAW,
			UPDATE,
			VISIT,
			DRAW,
			IMGUI,
			FRAME_END,

			COUNT
		};
	};

	/// Returns the configuration used to initialize the application
	inline const AppConfiguration &appConfiguration() const { return appCfg_; }
	/// Returns the run-time rendering settings
	inline RenderingSettings &renderingSettings() { return renderingSettings_; }
	/// Returns the debug overlay object, if any
	inline IDebugOverlay::DisplaySettings &debugOverlaySettings()
	{
		return (debugOverlay_) ? debugOverlay_->settings() : debugOverlayNullSettings_;
	}

	/// Returns all timings
	inline const float *timings() const { return timings_; }

	/// Returns the graphics device instance
	inline IGfxDevice &gfxDevice() { return *gfxDevice_; }
	/// Returns the root of the transformation graph
	inline SceneNode &rootNode() { return *rootNode_; }
	/// Returns the input manager instance
	inline IInputManager &inputManager() { return *inputManager_; }

	/// Returns the total number of frames already rendered
	unsigned long int numFrames() const;
	/// Returns the elapsed time since the end of the previous frame in milliseconds
	float interval() const;

	/// Returns the screen width as a float number
	inline float width() const { return static_cast<float>(gfxDevice_->width()); }
	/// Returns the screen height as a float number
	inline float height() const { return static_cast<float>(gfxDevice_->height()); }

	/// Returns the screen width as an integer number
	inline int widthInt() const { return gfxDevice_->width(); }
	/// Returns the screen height as an integer number
	inline int heightInt() const { return gfxDevice_->height(); }

	/// Returns the value of the suspension flag
	/*! If `true` the application is suspended, it will neither update nor receive events */
	inline bool isSuspended() const { return isSuspended_; }
	/// Sets the suspension flag value
	void setSuspended(bool suspended);

	/// Returns the value of the auto-suspension flag
	/*! If `true` the application will be suspended when it loses focus */
	inline bool autoSuspension() const { return autoSuspension_; }
	/// Sets the auto-suspension flag value
	inline void setAutoSuspension(bool autoSuspension) { autoSuspension_ = autoSuspension; }

	/// Raises the quit flag
	inline void quit() { shouldQuit_ = true; }
	/// Returns the quit flag value
	inline bool shouldQuit() const { return shouldQuit_; }

	/// Returns the focus flag value
	inline bool hasFocus() const { return hasFocus_; }

  protected:
	/// Maximum length for the information strings
	static const unsigned int MaxTextLength = 256;

	bool isSuspended_;
	bool autoSuspension_;
	bool hasFocus_;
	bool shouldQuit_;
	const AppConfiguration appCfg_;
	RenderingSettings renderingSettings_;
	float timings_[Timings::COUNT];
	IDebugOverlay::DisplaySettings debugOverlayNullSettings_;

	TimeStamp profileStartTime_;
	nctl::UniquePtr<FrameTimer> frameTimer_;
	nctl::UniquePtr<IGfxDevice> gfxDevice_;
	nctl::UniquePtr<RenderQueue> renderQueue_;
	nctl::UniquePtr<SceneNode> rootNode_;
	nctl::UniquePtr<IDebugOverlay> debugOverlay_;
	nctl::UniquePtr<IInputManager> inputManager_;
	nctl::UniquePtr<IAppEventHandler> appEventHandler_;
#ifdef WITH_IMGUI
	nctl::UniquePtr<ImGuiDrawing> imguiDrawing_;
#endif

	Application();
	~Application();

	/// Must be called before giving control to the application
	void initCommon();
	/// A single step of the game loop made to render a frame
	void step();
	/// Must be called before exiting to shut down the application
	void shutdownCommon();

	/// Sets the focus flag
	virtual void setFocus(bool hasFocus);

  private:
	/// Deleted copy constructor
	Application(const Application &) = delete;
	/// Deleted assignment operator
	Application &operator=(const Application &) = delete;

	bool shouldSuspend();

	friend class PCApplication;
	friend class AndroidApplication;
};

// Meyers' Singleton
extern DLL_PUBLIC Application &theApplication();

}

#endif
